#include <sys/types.h>
#include <utmp.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <pwd.h>
#include "untamo.h"

#ifdef PUCC
#include "rld.h"
#include "passfile.h"
#define PWIDLE	6	/*accounting bit for no idle time logout     */
#define PWMULT	7	/*accounting bit for multiple logins allowed */
#endif PUCC

#ifdef pdp11
#include <sys/param.h>
#define getdtablesize()	NOFILE
#endif pdp11

#ifdef PUCC
struct usrpwd *getupnam();
#endif PUCC

#include "y.tab.h"

struct user users[MAXUSERS];
struct user *pusers[MAXUSERS];

extern char *malloc(), *strcpy(), *ctime() , *strcat();
extern unsigned strlen();
extern time_t time();

struct qelem *rules, 
	     *session,
	     *exmpt;	/* lists for timeouts, session limits, and exemptions */

jmp_buf env_buf;
int sleeptime;		/* time to sleep between checks			  */
FILE *logfp;		/* log file file descriptor pointer		  */
int m_threshold;	/* number of users before multiple limits 	  */
int s_threshold;	/* number of users for session limits		  */
int warn_flags = IS_IDLE | IS_MULT | IS_LIMIT; 
			/* what sorts of warnings should be accepted    */
int fl_debug = 0, fl_tune = 0;

main(n_args, ppch_args)
	int n_args;
	char **ppch_args;
{
	struct utmp utmpbuf;
	struct stat statbuf;
#ifdef PUCC
	struct usrpwd *pswd;
#else PUCC
	struct passwd *pswd;
#endif PUCC
	struct user *user;
	char pathbuf[20];
	int utmptr, utmpfp;
	time_t conf_oldstamp;
	int userptr;
	int res, td;
	int new;		/* if the configuration file is new */
	time_t tempus;
	int finish(), wakeup();
	FILE *conffp, *freopen(), *fopen();
	static int fl_multiple = 0;
	/* command line flags */

	while( --n_args && *++ppch_args && **ppch_args == '-' ) {
		while( *++(*ppch_args) ){
			switch( **ppch_args ){
			case 'm': /* don't even think about multiple logins */
				warn_flags = warn_flags & (IS_IDLE|IS_LIMIT);
				break;
			case 'i': /* don't even think about idle timeouts */
				warn_flags = warn_flags & (IS_MULT|IS_LIMIT);
				break;
			case 's': /* don't even think about session limits */
				warn_flags = warn_flags & (IS_IDLE|IS_MULT);
				break;
			case 'd': /* set debug on */
				fl_debug = 1;
				fl_tune = 1;
				break;
			case 't': /* turn tuning option on */
				fl_tune = 1;
				break;
			default:
				fprintf( stderr, 
				"untamo: bad flag -%c\n", **ppch_args );
				break;
			}
		}
	}

	if( 0 == warn_flags ){
		/* do absolutely nothing!! */
		exit(0);
	}

	if ( access( "/flags/testsys" , F_OK ) == 0 ) {
		fprintf(stderr, "System in test mode\n");
		exit(0);			/* dont run in test mode */
	}
	(void) signal(SIGHUP,  SIG_IGN);	
	(void) signal(SIGQUIT, SIG_IGN);
	(void) signal(SIGINT,  SIG_IGN);
#ifndef pdp11
	(void) signal(SIGTTOU, SIG_IGN);
	(void) signal(SIGTSTP, SIG_IGN);
#endif pdp11

	(void) signal(SIGTERM, finish);
#ifdef pdp11
	if (1 == setjmp(env_buf)) {
#else  pdp11
	if (1 == _setjmp(env_buf)) {
#endif pdp11
	    fprintf(stderr, "Untamo recieved signal before startup\n");
	}
	(void) signal(SIGALRM, wakeup);
	conf_oldstamp = 1;			/* a very old stamp */
	/*
	 * set up new header nodes for each of the lists.
	 * The forw and back pointers must point to them
	 * selves so the system insque routine can be used
	 */
	rules = (struct qelem *) malloc( sizeof(struct qelem) );
	exmpt = (struct qelem *) malloc( sizeof(struct qelem) );
	session = (struct qelem *) malloc( sizeof(struct qelem) );
	rules->q_forw = rules->q_back = rules;
	exmpt->q_forw = exmpt->q_back = exmpt;
	session->q_forw = session->q_back = session;
	rules->q_item = session->q_item = exmpt->q_item = NULL;


	/*
	 * lose our controlling terminal
	 * Note:  We don't want to do this during debug so we can still get
	 * 	error messages
	 */
#ifndef DEBUG
	for (td = 0; td < getdtablesize(); td++)
		(void) close(td);
#endif DEBUG
	td = open("/dev/tty", O_RDWR, 0600);
	if (td >= 0){
		(void) ioctl(td, TIOCNOTTY, (char *)0);
		(void) close( td );
	}
	(void)open("/dev/null", O_RDONLY);  /* Give ourselves bogus I/O */
	(void)open("/dev/null", O_WRONLY);  /* for stdin/stdout/stderr so */
	(void)open("/dev/null", O_WRONLY);  /* our logfp doesn't get trashed */
	if ( (logfp = fopen(LOGFILE,"a")) > 0)  {
		if (-1 == flock(fileno(logfp), LOCK_EX | LOCK_NB)) {
			perror("No lock on logfile");
			exit(1);
		}
		Log("Untamo started");
	} else {
		(void) fprintf( stderr , "Untamo: couldn't open log file: %s\n", LOGFILE );
		(void) fflush(stderr);
		exit(1);
	}

	if ( (res = fork()) < 0)	{
		(void) fprintf(stderr,"Untamo: couldn't start\n");
		Log("Untamo: couldn't start");
		(void) fflush(stderr);
	}
	if (res){  /* if the parent */
		Log("Untamo PID (%d)",res);
		Log("session(%s) idle(%s) multiple(%s) debug(%s) tune(%s)",
			( (warn_flags & IS_LIMIT) ? "on" : "off" ),
			( (warn_flags & IS_IDLE) ? "on" : "off" ),
			( (warn_flags & IS_MULT) ? "on" : "off" ),
			( fl_debug ? "on" : "off" ),
			( fl_tune ? "on" : "off" ));
		exit(0);
	}

	/*
	 * now sit in an infinite loop and work
	 */

while (1){
#ifdef pdp11
	setjmp(env_buf);  /* return here if signal received */
#else  pdp11
	_setjmp(env_buf);  /* return here if signal received */
#endif pdp11
	if ( stat(CONFIG,&statbuf) < 0)  {
		Log("Untamo: couldn't stat conf file");
		exit(1);
	}

	if ( statbuf.st_mtime > conf_oldstamp ) {
		conf_oldstamp = statbuf.st_mtime;
		Log("Reading config file");

		if ( (conffp = freopen(CONFIG, "r", stdin)) < 0) {
			Log("Untamo: can't open configuration file");
			exit(1);
		}

		/*
		 * get rid of the old rules and exempt lists
		 */
		(void) freelist(rules);
		(void) freelist(exmpt);
		(void) freelist(session);
		m_threshold = 0;
		s_threshold = 0;
		/*
		 * now read the configuration file and set up the
		 * rules and exemption lists
		 */
		(void) yyparse();
		new = 1;
	} else
		new = 0;

	if ( (utmpfp = open(UTMP, O_RDONLY, 0)) < 0) {
		Log("Untamo: can't open /etc/utmp");
		exit(1);
	}

	utmptr = 0;
	userptr = 0;
	/*
	 * look through the utmp file, compare each entry to the users
	 * array to see if an entry has changed.  If it has, build a new
	 * record for that person, if it hasn't, see  if it is time to
	 * examine him again.
	 */
	while ( (res = read(utmpfp, (char *)&utmpbuf, sizeof(struct utmp)) ) > 0 ) {

		if (res != sizeof(struct utmp)) {
			Log("Untamo: error reading utmp file, continuing");
			continue;
		}
		(void) time(&tempus);
		if (utmpbuf.ut_name[0] != '\0')   {
			user = &users[utmptr];
			if ( !(strcmp(user->uid,utmpbuf.ut_name)) &&
			   (user->time_on == utmpbuf.ut_time) )	{
				if (new)
					(void) setlimits(utmptr);
				if ((warn_flags&IS_IDLE) && tempus >user->next){
					(void) checkidle(utmptr);
				}
			} else {
				/*
				 * build a new record
				 */
				user->warned  = 0;
				(void) strcpy(pathbuf,DEV);
				(void) strcat(pathbuf,utmpbuf.ut_line);
#ifdef PUCC
				user->rld = findrld(pathbuf);
				(void) strcpy(user->clust, findcluster(user->rld));
#endif PUCC

				(void) strcpy(user->line, pathbuf);
				(void) stat(pathbuf,&statbuf);
				(void) strcpy(user->uid, utmpbuf.ut_name);

#ifdef PUCC
				pswd = getupnam(utmpbuf.ut_name);
				user->ugroup = pswd->up_gid;
#else PUCC
				pswd = getpwnam(utmpbuf.ut_name);
				user->ugroup = pswd->pw_gid;
#endif PUCC

				user->time_on = utmpbuf.ut_time;
				(void) setlimits(utmptr);

#ifdef PUCC
				if( pswd->up_flags & (1l << PWMULT ))
				    user->exempt |= IS_MULT;
				if( pswd->up_flags & (1l << PWIDLE ))
				    user->exempt |= IS_IDLE;
#endif PUCC
				user->next = tempus;
 			}
			pusers[userptr++] = user;
		}
		utmptr++;
	}

	(void) close(utmpfp);
	(void) fclose(conffp);

	/*
	 * check session limits
	 * chk_session decides whether session limits are on inside
	 * itself.
	 */

	if( warn_flags & IS_LIMIT ){
		(void) chk_session(userptr);
	}

	/*
	 * check for and warn multiple logins
	 * when multiple logins are on and # users > threshold
	 */

	if((warn_flags & IS_MULT) && (userptr > m_threshold)){
		if ( 1 == fl_tune && 0 == fl_multiple ) {
			Log("multiple limits(on) users(%d) threshold(%d)",
				userptr, m_threshold);
			fl_multiple = 1;
		}
		(void) chk_multiple(userptr);
	} else {
		if (1 == fl_tune && 1 == fl_multiple ) {
			Log("multiple limits(off) users(%d) threshold(%d)",
				userptr, m_threshold);
			fl_multiple = 0;
		}
	}


	/*
	** wait sleeptime minutes
	*/

	(void) sleep( (unsigned) sleeptime * 60);
    }
}

/*
 * chk_session( users ) 
 * find out how many people are on sds ports, 
 * and try to warn enough people to get below the threshold
 */

chk_session( n_users )
	register int n_users;
{
	register int which_user;
	time_t tempus;
	register int n_ports = 0;
	static int fl_slimits = 0;

	(void) time(&tempus);
	for( which_user = 0; which_user < n_users ; which_user++ ){
		if( pusers[which_user]->warned & IS_LIMIT ){
			(void) warn(which_user,IS_LIMIT);
#ifdef CHKSDS
		} else if( is_sds_port( pusers[which_user]->rld ) ){
#else  CHKSDS
		} else {
#endif CHKSDS
			n_ports++;
		}
	}

	if( n_ports > s_threshold && !fl_slimits ){
		if ( 1 == fl_tune )
			Log("session limits(on) users(%d) threshold(%d)",
				n_ports, s_threshold);
		(void) close( creat( "/flags/sessionlimits", 0600 ));
		fl_slimits = 1;
	}

	if( n_ports < s_threshold && fl_slimits ){
		if ( 1 == fl_tune )
			Log("session limits(off) users(%d) threshold(%d)",
				n_ports, s_threshold);
		unlink( "/flags/sessionlimits" );
		fl_slimits = 0;
	}
		
	while( n_ports>s_threshold  && s_threshold>0 && which_user>0 ){
		which_user--;
#ifdef DEBUG
		Log("**debug: chksession(%s); session: %ld; exempt: %d", 
		    pusers[which_user]->uid,
		    pusers[which_user]->session,
		    pusers[which_user]->exempt);
#endif DEBUG
		if( tempus-pusers[which_user]->time_on > pusers[which_user]->session
		    && (pusers[which_user]->session > 2 * 60)
		    && !(pusers[which_user]->warned & IS_LIMIT)
		    && !(pusers[which_user]->exempt & IS_LIMIT)){

			(void) warn(which_user,IS_LIMIT);
			n_ports--;
		}
	}
}


#ifdef PUCC
int 
is_sds_port( rld_number )
	int rld_number;
{
	int fd, res;
	struct rld_data rdat;

	/*
	 * Get to the right place in /etc/rlds
	 * Complements of Jeff Smith...
	 */
	if( fd = open (RLD_FILE, O_RDONLY, 0) >= 0 ) {
		lseek (fd, (long) rld_number * sizeof (struct rld_data), L_SET);
		res = read (fd, (char *) &rdat, sizeof (struct rld_data));
		(void) close(fd);
		if (res == sizeof(struct rld_data) && rdat.rld_tio != -1) {
			return 1;
		}
	}
	return 0;
}
#endif PUCC

/*
 * chk_multiple -- given the number of users (i), warn any of 
 *	       them  that have multiple logins.  Calls qsort(3)
 *	       to sort them by id.
 */
chk_multiple(i)
int i;
{
	int j, comp(); 
	int match, skip = -1;
	int wait = 0;

	(void) qsort( (char *) pusers, i, sizeof(struct user *), comp);
	for (j=0; j<i-1; j++)	{

#ifdef DEBUG
		Log("**debug: chkmult(%s); exempt: %d", 
		    pusers[j]->uid,
		    pusers[j]->exempt);
#endif DEBUG

		/*
		 * if not all the multiple logins logged out,
		 * decide on one not to kill, clear his warned
		 * bit, and continue.  But don't look again until
		 * we have passed all the guys with the same login.
		 */
		if ( wait == 0 )  {
			match = 0;
			skip = decide(j, i, &wait);
		} else
			wait--;

		if ( ( pusers[j]->exempt & IS_MULT) || (j == skip) )  {
			continue;	/* he's exempt ! */
		}

		if ( !strcmp( pusers[j]->uid, pusers[j+1]->uid) )	{
			match = 1;
			(void) warn(j,IS_MULT);
		} else {
			if ( match )
				(void) warn(j,IS_MULT);
			match = 0;
		}
	}
}


/*
 * decide -- given a bunch of multiply logged on terminals that did
 * 	     not heed the warning, decide returns the index into the 
 *	     pusers array of the user NOT to log off.  Wait is the
 *  	     number of ids that chk_multiple must skip before calling
 *	     decide again.  Admittedly this is gross, but it works.
 */
decide(j, num, wait)
int j, num, *wait;
{
	int i;
	int count = 1;
	int warned = 1;
	int skip;

	/* 
	 * look through the users and find how many 
	 * of login pusers[i]->uid are logged on
	 * and whether or not they have been warned
	 */
	for ( i=j; i<num; i++)  {
		if ( !(pusers[i]->warned & IS_MULT) )
			warned = 0;

		if ( !strcmp( pusers[i]->uid, pusers[i+1]->uid) )
			count++;
		else
			break;
	}
	/*
	 * now, if there is a need to skip someone, do it
	 */
	*wait = count-1;
	if ( (warned) && (count > 1) )  {
		skip = j;
		/*
		 * set skip to the alpha-numerical least tty
		 */
		for(i=j+1; i<j+count; i++)
			if (strcmp(pusers[skip]->line, pusers[i]->line)>0)
				skip = i;
		pusers[skip]->warned &= ~IS_MULT;
		return(skip);
	}
	return(-1);
}

	
/*
 * finish -- end Untamo
 */
finish()
{
	(void) signal(SIGTERM, SIG_IGN);
	(void) unlink( "/flags/sessionlimits");
	Log("Untamo killed.");
	(void) flock(logfp, LOCK_UN);
#ifndef pdp11
	(void) fsync(logfp);
#endif  pdp11
	(void) fclose(logfp);
	exit(0);
}


/*
 * comp -- used by qsort to sort by id
 */
comp(h1, h2)
struct user **h1, **h2;
{
	return( strcmp((**h1).uid, (**h2).uid) );
}


/* 
 * checkidle -- given a user, see if we want to warn him about idle time.
 *		first check the exempt vector to see if he is exempt.
 */
#define min(a,b) ( (a)<(b)?(a):(b) )

checkidle(i)
int i;
{
	struct stat statbuf;
	time_t tempus;

	(void) time(&tempus);
	(void) stat(users[i].line,&statbuf);

#ifdef DEBUG
	Log("**debug: checkidle(%s); idle: %ld; exempt: %d",
	    users[i].uid, users[i].idle, users[i].exempt);
	Log(" *debug: %s %s", users[i].line, users[i].uid);
#endif DEBUG

	if (( tempus - statbuf.st_atime) < users[i].idle ) {
		users[i].warned &= ~IS_IDLE;
	} else {
		if (users[i].idle > 2*60 && !(users[i].exempt & IS_IDLE)) {
			(void) warn(i,IS_IDLE);
		}
	}
	users[i].next = min( statbuf.st_atime + users[i].idle,
			     users[i].time_on + users[i].session );
}

/*
 *=======================================================
 *
 *	Log(fmt, args) - Log a message in the untamo logfile
 *
 *	entry:
 *		fmt	- the format
 *		args	- the (optional) arguments
 *
 *=======================================================
 */

/*VARARGS1*/

Log(fmt, args)
char *fmt;
{

	time_t clock;             /* seconds from time() */
	static long oldclock;   /* clock time for last log entry */
	struct tm *localtime();
	struct tm *ptm;

	(void) time(&clock);
	if ( clock != oldclock ) {
		ptm = localtime(&clock);
		(void) fprintf(logfp, "%02d:%02d:%02d: ",  ptm->tm_hour,
							ptm->tm_min,
               		 				ptm->tm_sec);
	} else {
		(void) fprintf(logfp, "\t  ");
	}
	_doprnt(fmt, &args, logfp);
	(void) fprintf(logfp, "\n");
	(void) fflush(logfp);
	oldclock = clock;
}
